import asyncio
import gc
from unittest import mock

import pytest
import pyodbc

import aioodbc


def test_connect(loop, conn):
    assert conn.loop is loop
    assert not conn.autocommit
    assert conn.timeout == 0
    assert not conn.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_basic_cursor(conn):
    cursor = await conn.cursor()
    sql = 'SELECT 10;'
    await cursor.execute(sql)
    (resp,) = await cursor.fetchone()
    assert resp == 10


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_default_event_loop(loop, dsn):
    asyncio.set_event_loop(loop)
    conn = await aioodbc.connect(dsn=dsn)
    assert conn._loop is loop
    await conn.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_close_twice(conn):
    await conn.close()
    await conn.close()
    assert conn.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_execute(conn):
    cur = await conn.execute('SELECT 10;')
    (resp,) = await cur.fetchone()
    await conn.close()
    assert resp == 10
    assert conn.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_getinfo(conn):
    data = await conn.getinfo(pyodbc.SQL_CREATE_TABLE)
    pg = 14057
    sqlite = 1793
    mysql = 3093
    assert data in (pg, sqlite, mysql)


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_output_conversion(conn, table):
    def convert(value):
        # `value` will be a string.  We'll simply add an X at the
        # beginning at the end.
        return 'X' + value + 'X'

    await conn.add_output_converter(pyodbc.SQL_VARCHAR, convert)
    cur = await conn.cursor()

    await cur.execute("INSERT INTO t1 VALUES (3, '123.45')")
    await cur.execute("SELECT v FROM t1 WHERE n=3;")
    (value,) = await cur.fetchone()

    assert value == 'X123.45X'

    # Now clear the conversions and try again. There should be
    # no Xs this time.
    await conn.clear_output_converters()
    await cur.execute("SELECT v FROM t1")
    (value,) = await cur.fetchone()
    assert value == '123.45'
    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
async def test_autocommit(loop, connection_maker):
    conn = connection_maker(loop, autocommit=True)
    assert conn.autocommit, True


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_rollback(conn):
    assert not conn.autocommit

    cur = await conn.cursor()
    await cur.execute("CREATE TABLE t1(n INT, v VARCHAR(10));")

    await conn.commit()

    await cur.execute("INSERT INTO t1 VALUES (1, '123.45');")
    await cur.execute("SELECT v FROM t1")
    (value,) = await cur.fetchone()
    assert value == '123.45'

    await conn.rollback()
    await cur.execute("SELECT v FROM t1;")
    value = await cur.fetchone()
    assert value is None
    await cur.execute("DROP TABLE t1;")
    await conn.commit()

    await conn.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test___del__(loop, dsn, recwarn):
    conn = await aioodbc.connect(dsn=dsn, loop=loop)
    exc_handler = mock.Mock()
    loop.set_exception_handler(exc_handler)

    del conn
    gc.collect()
    w = recwarn.pop()
    assert issubclass(w.category, ResourceWarning)

    msg = {'connection': mock.ANY,  # conn was deleted
           'message': 'Unclosed connection'}
    if loop.get_debug():
        msg['source_traceback'] = mock.ANY
    exc_handler.assert_called_with(loop, msg)


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_custom_executor(loop, dsn, executor):
    conn = await aioodbc.connect(dsn=dsn, executor=executor, loop=loop)
    assert conn._executor is executor
    cur = await conn.execute('SELECT 10;')
    (resp,) = await cur.fetchone()
    await conn.close()
    assert resp == 10
    assert conn.closed


@pytest.mark.run_loop
async def test_dataSources(loop, executor):
    data = await aioodbc.dataSources(loop, executor)
    assert isinstance(data, dict)


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_connection_simple_with(loop, conn):
    assert not conn.closed
    async with conn:
        pass

    assert conn.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_connect_context_manager(loop, dsn):
    async with aioodbc.connect(dsn=dsn, loop=loop) as conn:
        assert not conn.closed
    assert conn.closed



import asyncio
from pyodbc import dataSources as _dataSources

from .connection import connect, Connection
from .pool import create_pool, Pool


__version__ = '0.0.2'
__all__ = ['connect', 'Connection', 'create_pool', 'Pool', 'dataSources']

(connect, Connection, create_pool, Pool)  # pyflakes


async def dataSources(loop=None, executor=None):
    """Returns a dictionary mapping available DSNs to their descriptions.

    :param loop: asyncio compatible event loop
    :param executor: instance of custom ThreadPoolExecutor, if not supplied
        default executor will be used
    :return dict: mapping of dsn to driver description
    """
    loop = loop or asyncio.get_event_loop()
    sources = await loop.run_in_executor(executor, _dataSources)
    return sources



import asyncio
import os
from concurrent.futures import ThreadPoolExecutor

import pytest
import aioodbc


@pytest.fixture
def loop(request):
    old_loop = asyncio.get_event_loop()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    def fin():
        loop.close()
        asyncio.set_event_loop(old_loop)

    request.addfinalizer(fin)
    return loop


@pytest.fixture
def executor(request):
    executor = ThreadPoolExecutor(max_workers=3)

    def fin():
        executor.shutdown()

    request.addfinalizer(fin)
    return executor


sqlite = 'Driver=SQLite;Database=sqlite.db,'
pg = ('Driver=PostgreSQL Unicode;'
      'Server=localhost;Port=5432;'
      'Database=aioodbc;Uid=aioodbc;'
      'Pwd=passwd;')
mysql = ('Driver=MySQL;Server=localhost;'
         'Database=aioodbc;User=root;'
         'Password=')


def pytest_namespace():
    return {'dsn_list': [sqlite, pg],
            'pg': pg, 'sqlite': sqlite, 'mysql': mysql}


@pytest.fixture
def dsn(request):
    conf = os.environ.get('DSN', 'Driver=SQLite;Database=sqlite.db')
    return conf


@pytest.fixture
def conn(request, loop, dsn):
    return _connect(loop, dsn, request.addfinalizer)


@pytest.fixture
def connection_maker(request, dsn):
    def f(loop, **kw):
        return _connect(loop, dsn, request.addfinalizer, **kw)
    return f


@pytest.fixture
def pool_maker(request):
    def f(loop, **kw):
        return _connect_pool(loop, request.addfinalizer, **kw)
    return f


def _connect_pool(loop, finalizer, **kw):
    pool = loop.run_until_complete(aioodbc.create_pool(loop=loop, **kw))

    def fin():
        pool.close()
        loop.run_until_complete(pool.wait_closed())

    finalizer(fin)
    return pool


@pytest.fixture
def pool(request, loop, dsn):
    return _connect_pool(loop, request.addfinalizer, dsn=dsn)


def _connect(loop, dsn, finalizer, **kw):
    conn = loop.run_until_complete(aioodbc.connect(dsn=dsn, loop=loop, **kw))

    def fin():
        loop.run_until_complete(conn.close())

    finalizer(fin)
    return conn


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name):
        item = pytest.Function(name, parent=collector)
        if 'run_loop' in item.keywords:
            return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    """
    Run asyncio marked test functions in an event loop instead of a normal
    function call.
    """
    if 'run_loop' in pyfuncitem.keywords:
        funcargs = pyfuncitem.funcargs
        loop = funcargs['loop']
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}

        if not asyncio.iscoroutinefunction(pyfuncitem.obj):
            func = asyncio.coroutine(pyfuncitem.obj)
        else:
            func = pyfuncitem.obj
        loop.run_until_complete(func(**testargs))
        return True


def pytest_runtest_setup(item):
    if 'run_loop' in item.keywords and 'loop' not in item.fixturenames:
        # inject an event loop fixture for all async tests
        item.fixturenames.append('loop')


@pytest.fixture
def table(request, conn, loop):

    async def go():
        cur = await conn.cursor()

        await cur.execute("CREATE TABLE t1(n INT, v VARCHAR(10));")
        await cur.execute("INSERT INTO t1 VALUES (1, '123.45');")
        await cur.execute("INSERT INTO t1 VALUES (2, 'foo');")
        await conn.commit()
        await cur.close()

    async def drop_table():
        cur = await conn.cursor()
        await cur.execute("DROP TABLE t1;")
        await cur.commit()
        await cur.close()

    def fin():
        loop.run_until_complete(drop_table())

    request.addfinalizer(fin)

    loop.run_until_complete(go())
    return 't1'



from collections.abc import Coroutine


class _ContextManager(Coroutine):

    __slots__ = ('_coro', '_obj')

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def send(self, value):
        return self._coro.send(value)

    def throw(self, typ, val=None, tb=None):
        if val is None:
            return self._coro.throw(typ)
        elif tb is None:
            return self._coro.throw(typ, val)
        else:
            return self._coro.throw(typ, val, tb)

    def close(self):
        return self._coro.close()

    @property
    def gi_frame(self):
        return self._coro.gi_frame

    @property
    def gi_running(self):
        return self._coro.gi_running

    @property
    def gi_code(self):
        return self._coro.gi_code

    def __next__(self):
        return self.send(None)

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


class _PoolContextManager(_ContextManager):
    async def __aexit__(self, exc_type, exc, tb):
        self._obj.close()
        await self._obj.wait_closed()
        self._obj = None


class _PoolConnectionContextManager(_ContextManager):

    __slots__ = ('_coro', '_conn', '_pool')

    def __init__(self, coro, pool):
        self._coro = coro
        self._conn = None
        self._pool = pool

    def __await__(self):
        self._pool = None
        return self._coro.__await__()

    async def __aenter__(self):
        self._conn = await self._coro
        return self._conn

    async def __aexit__(self, exc_type, exc, tb):
        await self._pool.release(self._conn)
        self._pool = None
        self._conn = None



# copied from aiopg
# https://github.com/aio-libs/aiopg/blob/master/aiopg/pool.py

import asyncio
import collections

from .connection import connect
from .utils import _PoolContextManager, _PoolConnectionContextManager

__all__ = ['create_pool', 'Pool']


def create_pool(minsize=10, maxsize=10, echo=False, loop=None,
                **kwargs):
    return _PoolContextManager(_create_pool(minsize=minsize, maxsize=maxsize,
                               echo=echo, loop=loop, **kwargs))


async def _create_pool(minsize=10, maxsize=10, echo=False, loop=None,
                       **kwargs):
    if loop is None:
        loop = asyncio.get_event_loop()

    pool = Pool(minsize=minsize, maxsize=maxsize, echo=echo, loop=loop,
                **kwargs)
    if minsize > 0:
        with (await pool._cond):
            await pool._fill_free_pool(False)
    return pool


class Pool(asyncio.AbstractServer):
    """Connection pool"""

    def __init__(self, minsize, maxsize, echo, loop, **kwargs):
        if minsize < 0:
            raise ValueError("minsize should be zero or greater")
        if maxsize < minsize:
            raise ValueError("maxsize should be not less than minsize")
        self._minsize = minsize
        self._loop = loop
        self._conn_kwargs = kwargs
        self._acquiring = 0
        self._free = collections.deque(maxlen=maxsize)
        self._cond = asyncio.Condition(loop=loop)
        self._used = set()
        self._closing = False
        self._closed = False
        self._echo = echo

    @property
    def echo(self):
        return self._echo

    @property
    def minsize(self):
        return self._minsize

    @property
    def maxsize(self):
        return self._free.maxlen

    @property
    def size(self):
        return self.freesize + len(self._used) + self._acquiring

    @property
    def freesize(self):
        return len(self._free)

    @property
    def closed(self):
        return self._closed

    async def clear(self):
        """Close all free connections in pool."""
        with (await self._cond):
            while self._free:
                conn = self._free.popleft()
                await conn.close()
            self._cond.notify()

    def close(self):
        """Close pool.

        Mark all pool connections to be closed on getting back to pool.
        Closed pool doesn't allow to acquire new connections.
        """
        if self._closed:
            return
        self._closing = True

    async def wait_closed(self):
        """Wait for closing all pool's connections."""

        if self._closed:
            return
        if not self._closing:
            raise RuntimeError(".wait_closed() should be called "
                               "after .close()")

        while self._free:
            conn = self._free.popleft()
            await conn.close()

        with (await self._cond):
            while self.size > self.freesize:
                await self._cond.wait()

        self._closed = True

    def acquire(self):
        """Acquire free connection from the pool."""
        coro = self._acquire()
        return _PoolConnectionContextManager(coro, self)

    async def _acquire(self):
        if self._closing:
            raise RuntimeError("Cannot acquire connection after closing pool")
        with (await self._cond):
            while True:
                await self._fill_free_pool(True)
                if self._free:
                    conn = self._free.popleft()
                    assert not conn.closed, conn
                    assert conn not in self._used, (conn, self._used)
                    self._used.add(conn)
                    return conn
                else:
                    await self._cond.wait()

    async def _fill_free_pool(self, override_min):
        while self.size < self.minsize:
            self._acquiring += 1
            try:
                conn = await connect(echo=self._echo, loop=self._loop,
                                     **self._conn_kwargs)
                # raise exception if pool is closing
                self._free.append(conn)
                self._cond.notify()
            finally:
                self._acquiring -= 1
        if self._free:
            return

        if override_min and self.size < self.maxsize:
            self._acquiring += 1
            try:
                conn = await connect(echo=self._echo, loop=self._loop,
                                     **self._conn_kwargs)
                # raise exception if pool is closing
                self._free.append(conn)
                self._cond.notify()
            finally:
                self._acquiring -= 1

    async def _wakeup(self):
        with (await self._cond):
            self._cond.notify()

    async def release(self, conn):
        """Release free connection back to the connection pool.
        """
        assert conn in self._used, (conn, self._used)
        self._used.remove(conn)
        if not conn.closed:
            if self._closing:
                await conn.close()
            else:
                self._free.append(conn)
            await self._wakeup()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()
        await self.wait_closed()



import pyodbc
import pytest


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_cursor_with(loop, conn, table):
    ret = []

    # regular cursor usage
    cur = await conn.cursor()
    await cur.execute('SELECT * FROM t1;')
    assert not cur.closed

    # cursor should be closed
    async with cur:
        async for i in cur:
            ret.append(i)
    expected = [tuple(r) for r in ret]
    assert [(1, '123.45'), (2, 'foo')] == expected
    assert cur.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_cursor_lightweight(loop, conn, table):

    cur = await conn.cursor()
    await cur.execute('SELECT * FROM t1;')

    assert not cur.closed
    async with cur:
        pass

    assert cur.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_cursor_await(loop, conn, table):

    async with conn.cursor() as cur:
        await cur.execute('SELECT * FROM t1;')
        assert not cur.closed

    assert cur.closed


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_cursor(conn):
    cur = await conn.cursor()
    assert cur.connection is conn
    assert cur._loop, conn.loop
    assert cur.arraysize == 1
    assert cur.rowcount == -1

    r = await cur.setinputsizes()
    assert r is None

    await cur.setoutputsize()
    assert r is None
    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_execute_on_closed_cursor(conn):
    cur = await conn.cursor()
    await cur.close()
    with pytest.raises(pyodbc.OperationalError):
        await cur.execute('SELECT 1;')


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_close(conn):
    cur = await conn.cursor()
    assert not cur.closed
    await cur.close()
    await cur.close()
    assert cur.closed


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_description(conn):
    cur = await conn.cursor()
    assert cur.description is None
    await cur.execute('SELECT 1;')
    expected = (('1', float, None, 54, 54, 0, True), )
    assert cur.description == expected
    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_description_with_real_table(conn, table):
    cur = await conn.cursor()
    await cur.execute("SELECT * FROM t1;")

    expected = (('n', int, None, 10, 10, 0, True),
                ('v', str, None, 10, 10, 0, True))
    assert cur.description == expected
    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_rowcount_with_table(conn, table):
    cur = await conn.cursor()
    await cur.execute("SELECT * FROM t1;")
    await cur.fetchall()
    # sqlite does not provide working rowcount attribute
    # http://stackoverflow.com/questions/4911404/in-pythons-sqlite3-
    # module-why-cant-cursor-rowcount-tell-me-the-number-of-ro
    # TODO: figure out for proper test
    assert cur.rowcount in (0, 2)
    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_arraysize(conn):
    cur = await conn.cursor()
    assert 1 == cur.arraysize
    cur.arraysize = 10
    assert 10 == cur.arraysize
    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_fetchall(conn, table):
    cur = await conn.cursor()
    await cur.execute("SELECT * FROM t1;")
    resp = await cur.fetchall()
    expected = [(1, '123.45'), (2, 'foo')]

    for row, exp in zip(resp, expected):
        assert exp == tuple(row)

    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_fetchmany(conn, table):
    cur = await conn.cursor()
    await cur.execute("SELECT * FROM t1;")
    resp = await cur.fetchmany(1)
    expected = [(1, '123.45')]

    for row, exp in zip(resp, expected):
        assert exp == tuple(row)

    await cur.close()


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_fetchone(conn, table):
    cur = await conn.cursor()
    await cur.execute("SELECT * FROM t1;")
    resp = await cur.fetchone()
    expected = (1, '123.45')

    assert expected == tuple(resp)
    await cur.close()


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_tables(conn, table):
    cur = await conn.cursor()
    await cur.tables()
    resp = await cur.fetchall()
    expectd = (None, None, 't1', 'TABLE', None)
    assert len(resp) == 1, resp
    assert expectd == tuple(resp[0]), resp


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_cursor_rollback(conn, table):

    cur = await conn.cursor()
    await cur.execute("INSERT INTO t1 VALUES (3, '123.45');")
    await cur.execute("SELECT v FROM t1 WHERE n=3;")
    (value, ) = await cur.fetchone()
    assert value == '123.45'

    await cur.rollback()
    await cur.execute("SELECT v FROM t1 WHERE n=3;")
    value = await cur.fetchone()
    assert value is None


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_columns(conn, table):
    cur = await conn.cursor()
    await cur.columns()
    resp = await cur.fetchall()
    expectd = [('', '', 't1', 'n', 4, 'INT', 9, 10, 10, 0, 1, None,
                'NULL', 4, None, 16384, 1, 'YES'),
               ('', '', 't1', 'v', 12, 'VARCHAR(10)', 10, 10, 10, 0, 1, None,
                'NULL', 12, None, 16384, 2, 'YES')]
    columns = [tuple(r) for r in resp]
    assert expectd == columns


@pytest.mark.parametrize('dsn', pytest.dsn_list)
@pytest.mark.run_loop
async def test_executemany(conn):
    cur = await conn.cursor()
    await cur.execute("CREATE TABLE t1(a int, b VARCHAR(10))")
    # TODO: figure out why it is possible to insert only strings... but not int
    params = [(str(i), str(i)) for i in range(1, 6)]
    await cur.executemany("INSERT INTO t1(a, b) VALUES (?, ?)", params)
    await cur.execute("SELECT COUNT(*) FROM t1")
    count = await cur.fetchone()
    assert count[0] == len(params)

    await cur.execute("SELECT a, b FROM t1 ORDER BY a")
    rows = await cur.fetchall()
    assert count[0] == len(rows)

    for param, row in zip(params, rows):
        assert int(param[0]) == row[0]
        assert param[1] == row[1]
    await cur.execute("DROP TABLE t1;")


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_procedures_empty(conn, table):
    cur = await conn.cursor()
    await cur.procedures()
    resp = await cur.fetchall()
    assert resp == []


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_procedureColumns_empty(conn, table):
    cur = await conn.cursor()
    await cur.procedureColumns()
    resp = await cur.fetchall()
    assert resp == []


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_primaryKeys_empty(conn, table):
    cur = await conn.cursor()
    await cur.primaryKeys('t1', 't1', 't1')
    resp = await cur.fetchall()
    assert resp == []


@pytest.mark.parametrize('dsn', [pytest.sqlite])
@pytest.mark.run_loop
async def test_foreignKeys_empty(conn, table):
    cur = await conn.cursor()
    await cur.foreignKeys('t1')
    resp = await cur.fetchall()
    assert resp == []


@pytest.mark.run_loop
async def test_getTypeInfo_empty(conn, table):
    cur = await conn.cursor()
    await cur.getTypeInfo(pyodbc.SQL_CHAR)
    resp = await cur.fetchall()
    expected = [('char', 1, 255, "'", "'", 'length', 1, 0, 3, None, 0, 0,
                 'char', None, None, 1, 0, None, None)]
    type_info = [tuple(r) for r in resp]
    assert type_info == expected



import asyncio

import pytest
import aioodbc
from aioodbc import Pool, Connection


def test_create_pool(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn)
    assert isinstance(pool, Pool)
    assert 10 == pool.minsize
    assert 10 == pool.maxsize
    assert 10 == pool.size
    assert 10 == pool.freesize
    assert not pool.echo


def test_create_pool2(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, maxsize=20)
    assert isinstance(pool, Pool)
    assert 10 == pool.minsize
    assert 20 == pool.maxsize
    assert 10 == pool.size
    assert 10 == pool.freesize


@pytest.mark.parametrize('dsn', pytest.dsn_list)
def test_acquire(loop, pool):
    async def go():
        conn = await pool.acquire()
        try:
            assert isinstance(conn, Connection)
            assert not conn.closed
            cur = await conn.cursor()
            await cur.execute('SELECT 1')
            val = await cur.fetchone()
            assert (1,) == tuple(val)
        finally:
            await pool.release(conn)

    loop.run_until_complete(go())


def test_release(loop, pool):
    async def go():
        conn = await pool.acquire()
        try:
            assert 9 == pool.freesize
            assert {conn} == pool._used
        finally:
            await pool.release(conn)
        assert 10 == pool.freesize
        assert not pool._used

    loop.run_until_complete(go())


def test_release_closed(loop, pool):
    async def go():
        conn = await pool.acquire()
        assert 9 == pool.freesize
        await conn.close()
        await pool.release(conn)
        assert 9 == pool.freesize
        assert not pool._used
        assert 9 == pool.size

        conn2 = await pool.acquire()
        assert 9 == pool.freesize
        assert 10 == pool.size
        await pool.release(conn2)

    loop.run_until_complete(go())


def test_context_manager(loop, pool):

    async def go():
        conn = await pool.acquire()
        try:
            assert isinstance(conn, Connection)
            assert 9 == pool.freesize
            assert {conn} == pool._used
        finally:
            await pool.release(conn)
        assert 10 == pool.freesize

    loop.run_until_complete(go())


def test_clear(loop, pool):
    async def go():
        await pool.clear()
        assert 0 == pool.freesize

    loop.run_until_complete(go())


def test_initial_empty(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=0)

    async def go():
        assert 10 == pool.maxsize
        assert 0 == pool.minsize
        assert 0 == pool.size
        assert 0 == pool.freesize

        conn = await pool.acquire()
        try:
            assert 1 == pool.size
            assert 0 == pool.freesize
        finally:
            await pool.release(conn)
        assert 1 == pool.size
        assert 1 == pool.freesize

        conn1 = await pool.acquire()
        assert 1 == pool.size
        assert 0 == pool.freesize

        conn2 = await pool.acquire()
        assert 2 == pool.size
        assert 0 == pool.freesize

        await pool.release(conn1)
        assert 2 == pool.size
        assert 1 == pool.freesize

        await pool.release(conn2)
        assert 2 == pool.size
        assert 2 == pool.freesize

    loop.run_until_complete(go())


def test_parallel_tasks(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=0, maxsize=2)

    async def go():
        assert 2 == pool.maxsize
        assert 0 == pool.minsize
        assert 0 == pool.size
        assert 0 == pool.freesize

        fut1 = pool.acquire()
        fut2 = pool.acquire()

        conn1, conn2 = await asyncio.gather(fut1, fut2, loop=loop)
        assert 2 == pool.size
        assert 0 == pool.freesize
        assert {conn1, conn2} == pool._used

        await pool.release(conn1)
        assert 2 == pool.size
        assert 1 == pool.freesize
        assert {conn2} == pool._used

        await pool.release(conn2)
        assert 2 == pool.size
        assert 2 == pool.freesize
        assert not conn1.closed
        assert not conn2.closed

        conn3 = await pool.acquire()
        assert conn3 is conn1
        await pool.release(conn3)

    loop.run_until_complete(go())


def test_parallel_tasks_more(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=0, maxsize=3)

    async def go():
        fut1 = pool.acquire()
        fut2 = pool.acquire()
        fut3 = pool.acquire()

        conn1, conn2, conn3 = await asyncio.gather(fut1, fut2, fut3,
                                                   loop=loop)
        assert 3 == pool.size
        assert 0 == pool.freesize
        assert {conn1, conn2, conn3} == pool._used

        await pool.release(conn1)
        assert 3 == pool.size
        assert 1 == pool.freesize
        assert {conn2, conn3} == pool._used

        await pool.release(conn2)
        assert 3 == pool.size
        assert 2 == pool.freesize
        assert {conn3} == pool._used
        assert not conn1.closed
        assert not conn2.closed

        await pool.release(conn3)
        assert 3 == pool.size
        assert 3 == pool.freesize
        assert not pool._used
        assert not conn1.closed
        assert not conn2.closed
        assert not conn3.closed

        conn4 = await pool.acquire()
        assert conn4 is conn1
        await pool.release(conn4)

        loop.run_until_complete(go())


def test_default_event_loop(loop, dsn):
    asyncio.set_event_loop(loop)

    async def go():
        pool = await aioodbc.create_pool(dsn=dsn)
        assert pool._loop is loop
        pool.close()
        await pool.wait_closed()

    loop.run_until_complete(go())


def test__fill_free(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=1)

    async def go():
        first_conn = await pool.acquire()
        try:
            assert 0 == pool.freesize
            assert 1 == pool.size

            conn = await asyncio.wait_for(pool.acquire(), timeout=0.5,
                                          loop=loop)
            assert 0 == pool.freesize
            assert 2 == pool.size
            await pool.release(conn)
            assert 1 == pool.freesize
            assert 2 == pool.size
        finally:
            await pool.release(first_conn)
        assert 2 == pool.freesize
        assert 2 == pool.size

    loop.run_until_complete(go())


def test_connect_from_acquire(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=0)

    async def go():
        assert 0 == pool.freesize
        assert 0 == pool.size
        conn = await pool.acquire()
        try:
            assert 1 == pool.size
            assert 0 == pool.freesize
        finally:
            await pool.release(conn)
        assert 1 == pool.size
        assert 1 == pool.freesize

    loop.run_until_complete(go())


def test_concurrency(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=2, maxsize=4)

    async def go():
        c1 = await pool.acquire()
        c2 = await pool.acquire()
        assert 0 == pool.freesize
        assert 2 == pool.size
        await pool.release(c1)
        await pool.release(c2)

    loop.run_until_complete(go())


def test_invalid_minsize_and_maxsize(loop, dsn):
    async def go():
        with pytest.raises(ValueError):
            await aioodbc.create_pool(dsn=dsn, loop=loop, minsize=-1)

        with pytest.raises(ValueError):
            await aioodbc.create_pool(dsn=dsn, loop=loop, minsize=5,
                                      maxsize=2)

    loop.run_until_complete(go())


def test_true_parallel_tasks(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, minsize=0, maxsize=1)

    async def go():
        assert 1 == pool.maxsize
        assert 0 == pool.minsize
        assert 0 == pool.size
        assert 0 == pool.freesize

        maxsize = 0
        minfreesize = 100

        async def inner():
            nonlocal maxsize, minfreesize
            maxsize = max(maxsize, pool.size)
            minfreesize = min(minfreesize, pool.freesize)
            conn = await pool.acquire()
            maxsize = max(maxsize, pool.size)
            minfreesize = min(minfreesize, pool.freesize)
            await asyncio.sleep(0.01, loop=loop)
            await pool.release(conn)
            maxsize = max(maxsize, pool.size)
            minfreesize = min(minfreesize, pool.freesize)

        await asyncio.gather(inner(), inner(), loop=loop)

        assert 1 == maxsize
        assert 0 == minfreesize

    loop.run_until_complete(go())


def test_cannot_acquire_after_closing(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn)

    async def go():
        pool.close()

        with pytest.raises(RuntimeError):
            await pool.acquire()

    loop.run_until_complete(go())


def test_wait_closed(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn)

    async def go():
        c1 = await pool.acquire()
        c2 = await pool.acquire()
        assert 10 == pool.size
        assert 8 == pool.freesize

        ops = []

        async def do_release(conn):
            await asyncio.sleep(0, loop=loop)
            await pool.release(conn)
            ops.append('release')

        async def wait_closed():
            await pool.wait_closed()
            ops.append('wait_closed')

        pool.close()
        await asyncio.gather(wait_closed(),
                             do_release(c1),
                             do_release(c2),
                             loop=loop)
        assert ['release', 'release', 'wait_closed'] == ops
        assert 0 == pool.freesize

    loop.run_until_complete(go())


def test_echo(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn, echo=True)

    async def go():
        assert pool.echo
        conn = await pool.acquire()
        assert conn.echo
        await pool.release(conn)

    loop.run_until_complete(go())


def test_release_closed_connection(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn)

    async def go():
        conn = await pool.acquire()
        await conn.close()

        await pool.release(conn)
        pool.close()

    loop.run_until_complete(go())


def test_wait_closing_on_not_closed(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn)

    async def go():
        with pytest.raises(RuntimeError):
            await pool.wait_closed()
        pool.close()

    loop.run_until_complete(go())


def test_close_with_acquired_connections(loop, pool_maker, dsn):
    pool = pool_maker(loop, dsn=dsn)

    async def go():
        conn = await pool.acquire()
        pool.close()

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(pool.wait_closed(), 0.1, loop=loop)
        await conn.close()
        await pool.release(conn)

    loop.run_until_complete(go())


@pytest.mark.parametrize('dsn', pytest.dsn_list)
def test_pool_with_executor(loop, pool_maker, dsn, executor):
    pool = pool_maker(loop, executor=executor, dsn=dsn, minsize=2, maxsize=2)

    async def go():
        conn = await pool.acquire()
        try:
            assert isinstance(conn, Connection)
            assert not conn.closed
            assert conn._executor is executor
            cur = await conn.cursor()
            await cur.execute('SELECT 1')
            val = await cur.fetchone()
            assert (1,) == tuple(val)
        finally:
            await pool.release(conn)
        # we close pool here instead in finalizer because of pool should be
        # closed before executor
        pool.close()
        await pool.wait_closed()

    loop.run_until_complete(go())


@pytest.mark.parametrize('dsn', pytest.dsn_list)
def test_pool_context_manager(loop, pool):
    async def go():
        assert not pool.closed
        async with pool:
            assert not pool.closed
        assert pool.closed

    loop.run_until_complete(go())


@pytest.mark.parametrize('dsn', pytest.dsn_list)
def test_pool_context_manager2(loop, pool):
    async def go():
        async with pool.acquire() as conn:
            assert not conn.closed
            cur = await conn.cursor()
            await cur.execute('SELECT 1')
            val = await cur.fetchone()
            assert (1,) == tuple(val)

    loop.run_until_complete(go())


@pytest.mark.parametrize('dsn', pytest.dsn_list)
def test_all_context_managets(dsn, loop):
    async def go():
        async with aioodbc.create_pool(dsn=dsn, loop=loop) as pool:
            async with pool.acquire() as conn:
                async with conn.cursor() as cur:
                    assert not pool.closed
                    assert not conn.closed
                    assert not cur.closed

                    await cur.execute('SELECT 1')
                    val = await cur.fetchone()
                    assert (1,) == tuple(val)

        assert pool.closed
        assert conn.closed
        assert cur.closed
    loop.run_until_complete(go())
